# -*- coding: utf-8 -*-
"""
Created on Thu Nov  8 09:29:34 2018

@author:  audiffren
"""

import numpy as np
import saved_values as SAVE  # this file contains the values for the GMM


# Some utils GMM related functions
def _log_multivariate_normal_density(X, means, covars):
    """Compute Gaussian log-density at X for a diagonal covariance model,
    simultaneously for the two clusters and all the windows"""
    n_samples, n_dim = X.shape
    log_proba = -0.5 *( \
                 (n_dim * np.log(2 * np.pi)) +\
                 np.sum(np.log(covars[None,:,:]), axis=2) +\
                 np.sum( ((X[:,None,:] - means[None,:,:])**2)/covars[None,:,:],axis=2 ) )
    return log_proba

def _logsumexp(arr, axis=1):
    """  from log probability of each gaussian, deduce log of the sum of the proba"""
    return np.log(np.sum(np.exp(arr),axis=1))





# Features
# This part of the code compute the different features on each window, which then serve
# as the GMM input

def velocity(signal):
    #This compute the median norm of the displacement of the CoP
    difs = np.diff(signal,axis=0,n=1)
    spd = np.linalg.norm(difs,axis=1)
    return np.median(spd)


def simplified_area(signal):
    # This computes an approximation of the statokinesigram area

    m = np.mean(signal,axis=0, keepdims=True)
    conv = (signal - m).T @ (signal - m) / (len(signal)-1)

    det = np.linalg.det(conv)
    return 1.96 * np.sqrt(det)

def ml_range(signal):
    # This compute the width (medio laterak axis) of the statokinesigram
    return max(signal[:,0]) - min(signal[:,0])




def compute_features(signal):
    features=[]
    features.append(ml_range(signal))
    features.append(velocity(signal))
    features.append(simplified_area(signal))
    return features





class GMM_Score_predictor :
    """
    This class relies on a GMM model to predict the score of the successive window of a trajectory
    Its parameter is the block size in seconds, to be chosen among [1,2,3], and the eye type among
    [OE,CE]  which stand for OpenEyes, CloseEyes

    """

    def __init__(self,block_size,eye_type):
        self.eye_type = eye_type
        # this variable contains the windows length corresponding to the block size
        self.window=25*block_size
        #this defines the overlap between windows
        self.overlap=int(25//2)

        # means, stds of the features for normalization purpose
        self.means=SAVE.MEANS[block_size][eye_type]
        self.stds= SAVE.STDS[block_size][eye_type]

        # means, covars of the gaussians associated with each cluster

        self.gmms_means = SAVE.GMM_CENTERS[block_size][eye_type]
        self.gmms_covar = SAVE.GMM_COVARS[block_size][eye_type]

        # weights of each cluster
        self.gmms_bias = SAVE.GMM_CLUSTER_WEIGTHS[block_size][eye_type]

        # index of the cluster corresponding to the quiet phase
        self.gmms_reverse = SAVE.GMM_QUIET_CLUSTER_INDEX[block_size][eye_type]



    def _prepare_windows_one(self,signal):
        """
        This function compute the features for each window of a given signal.
        Those features are then normalized.
        """
        feats=[]
        # for each window
        for winstart in range(0,len(signal)-self.window,self.overlap):
            # we select the relevant part of the signal
            win=signal[winstart:winstart+self.window]
            # we compute the features of this window
            f =compute_features(win)
            # the features are normalized
            f-= self.means
            f/= self.stds
            feats.append(f)
        return feats


    def _compute_gmm_proba(self,X,quiet_cluster_index=None):
        """
        This function compute the probability for each window's feature
        to belong to the 'quiet' cluster
        """
        # this vector contains for each window, the log of the probaility of each gaussian (one gaussian per cluster)
        log_proba_gaussian = _log_multivariate_normal_density(X,  means= self.gmms_means,
                                               covars = self.gmms_covar)
        # we add the log of the weights of the clusters
        log_proba_gaussian += np.log(self.gmms_bias[None,:])

        # we compute the log of the sum of the proba of the two gaussians
        log_total_proba = _logsumexp(log_proba_gaussian, axis=1)

        # we deduce the proba of belonging to each cluster
        responsibilities = np.exp(log_proba_gaussian - log_total_proba[:, None])

        quiet_cluster_proba = responsibilities[:,quiet_cluster_index]
        return quiet_cluster_proba





    def predict(self,signal):
        """
        predict the score of a stato
        """
        features_on_windows = self._prepare_windows_one(signal) # compute the features on each window
        score_predictions = self._compute_gmm_proba(np.array(features_on_windows),self.gmms_reverse) # compute the score on each window
        return score_predictions







